from concurrent.futures import ProcessPoolExecutor, as_completed
from numba import njit
from joblib import Memory
import numpy as np

# from pegasus.Ocean import ImageBindEmbeddingFunctio
from pegasus.oceandb import ImageBindEmbeddingFunction

memory = Memory("PegasusStore", verbose=0)

@memory.cache
@njit
def optimized_embedding_function(modality, data):
    return ImageBindEmbeddingFunction(modality)(data)

class Pegasus:
    def __init__(self, modality, multi_process=False, n_processes=4):
        self.modality = modality
        self.multi_process = multi_process
        self.n_processes = n_processes if multi_process else 1
        
    def _embed_data(self, data):
        if self.modality not in {"text", "audio", "vision", "sensor", "heatmap"}:
            raise ValueError("Invalid modality")
        return optimized_embedding_function(self.modality, data)
    
    def embed_data(self, data):
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        
        if not self.multi_process:
            return self._embed_data(data)
        
        with ProcessPoolExecutor(max_workers=self.n_processes) as executor:
            future_to_data = {executor.submit(self._embed_data, d): d for d in data}
            return {future_to_data[future]: future.result() for future in as_completed(future_to_data)}


from abc import abstractmethod


class OceanError(Exception):
    def code(self):
        """Return an appropriate HTTP response code for this error"""
        return 400  # Bad Request

    def message(self):
        return ", ".join(self.args)

    @classmethod
    @abstractmethod
    def name(self):
        """Return the error name"""
        pass


class NoDatapointsException(OceanError):
    @classmethod
    def name(cls):
        return "NoDatapoints"


class NoIndexException(OceanError):
    @classmethod
    def name(cls):
        return "NoIndex"


class InvalidDimensionException(OceanError):
    @classmethod
    def name(cls):
        return "InvalidDimension"


class NotEnoughElementsException(OceanError):
    @classmethod
    def name(cls):
        return "NotEnoughElements"


class IDAlreadyExistsError(OceanError):
    def code(self):
        return 409  # Conflict

    @classmethod
    def name(cls):
        return "IDAlreadyExists"


class DuplicateIDError(OceanError):
    @classmethod
    def name(cls):
        return "DuplicateID"


class InvalidUUIDError(OceanError):
    @classmethod
    def name(cls):
        return "InvalidUUID"


error_types = {
    "NoDatapoints": NoDatapointsException,
    "NoIndex": NoIndexException,
    "InvalidDimension": InvalidDimensionException,
    "NotEnoughElements": NotEnoughElementsException,
    "IDAlreadyExists": IDAlreadyExistsError,
    "DuplicateID": DuplicateIDError,
    "InvalidUUID": InvalidUUIDError,
}



from oceandb.api.types import Documents, EmbeddingFunction, Embeddings
from typing import Optional
import torch

from .ImageBind.models import imagebind_model
from .ImageBind.models.imagebind_model import ModalityType
from .ImageBind.data import load_and_transform_text, load_and_transform_vision_data, load_and_transform_audio_data


class ImageBindEmbeddingFunction(EmbeddingFunction):
    def __init__(
        self,
        modality: str = ModalityType,  # type: ignore
        model_path: str = "https://dl.fbaipublicfiles.com/imagebind/imagebind_huge.pth",
        device: str = "cuda:0"
    ):
        self._modality = modality
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self._model = imagebind_model.imagebind_huge(pretrained=True)
        self._model.eval()
        self._model.to(self.device)

    def __call__(self, *args: Documents) -> Embeddings:
        if self._modality == ModalityType.TEXT:
            inputs = {
                ModalityType.TEXT: load_and_transform_text(
                    args[0], self.device
                )
            }
            print("Inputs:", inputs)
        elif self._modality == ModalityType.VISION:
            inputs = {
                ModalityType.VISION: load_and_transform_vision_data(
                    args[0], self.device
                )
            }
        elif self._modality == ModalityType.AUDIO:
            inputs = {
                ModalityType.AUDIO: load_and_transform_audio_data(
                    args[0], self.device
                )
            }
        else:
            raise ValueError("Invalid modality specified")

        with torch.no_grad():
            embeddings = self._model(inputs)
        
        print("Embeddings:", embeddings)

        # Convert the embeddings tensor to a NumPy array and then to a list of lists (embeddings)
        embeddings_array = embeddings[self._modality].cpu().numpy()

        print("Embeddings array:", embeddings_array)
        # embeddings_list = embeddings_array.tolist()

        # return embeddings_list

        return [embedding.tolist() for embedding in embeddings_array]


"""
text_embedding_function = ImageBindEmbeddingFunction(modality=ModalityType.TEXT)
vision_embedding_function = ImageBindEmbeddingFunction(modality=ModalityType.VISION)
audio_embedding_function = ImageBindEmbeddingFunction(modality=ModalityType.AUDIO)


"""


class OpenAIEmbeddingFunction(EmbeddingFunction):
    def __init__(
        self, api_key: Optional[str] = None, model_name: str = "text-embedding-ada-002"
    ):
        try:
            import openai
        except ImportError:
            raise ValueError(
                "The openai python package is not installed. Please install it with `pip install openai`"
            )

        if api_key is not None:
            openai.api_key = api_key
        # If the api key is still not set, raise an error
        elif openai.api_key is None:
            raise ValueError(
                "Please provide an OpenAI API key. You can get one at https://platform.openai.com/account/api-keys"
            )

        self._client = openai.Embedding
        self._model_name = model_name

    def __call__(self, texts: Documents) -> Embeddings:
        # replace newlines, which can negatively affect performance.
        texts = [t.replace("\n", " ") for t in texts]

        # Call the OpenAI Embedding API
        embeddings = self._client.create(input=texts, engine=self._model_name)["data"]

        # Sort resulting embeddings by index
        sorted_embeddings = sorted(embeddings, key=lambda e: e["index"])

        # Return just the embeddings
        return [result["embedding"] for result in sorted_embeddings]


class CohereEmbeddingFunction(EmbeddingFunction):
    def __init__(self, api_key: str, model_name: str = "large"):
        try:
            import cohere
        except ImportError:
            raise ValueError(
                "The cohere python package is not installed. Please install it with `pip install cohere`"
            )

        self._client = cohere.Client(api_key)
        self._model_name = model_name

    def __call__(self, texts: Documents) -> Embeddings:
        # Call Cohere Embedding API for each document.
        return [
            embeddings
            for embeddings in self._client.embed(texts=texts, model=self._model_name)
        ]


class HuggingFaceEmbeddingFunction(EmbeddingFunction):
    def __init__(
        self, api_key: str, model_name: str = "sentence-transformers/all-MiniLM-L6-v2"
    ):
        try:
            import requests
        except ImportError:
            raise ValueError(
                "The requests python package is not installed. Please install it with `pip install requests`"
            )
        self._api_url = f"https://api-inference.huggingface.co/pipeline/feature-extraction/{model_name}"
        self._session = requests.Session()
        self._session.headers.update({"Authorization": f"Bearer {api_key}"})

    def __call__(self, texts: Documents) -> Embeddings:
        # Call HuggingFace Embedding API for each document
        return self._session.post(
            self._api_url, json={"inputs": texts, "options": {"wait_for_model": True}}
        ).json()




from pydantic import BaseSettings
from typing import List

TELEMETRY_WHITELISTED_SETTINGS = [
    "ocean_db_impl",
    "ocean_api_impl",
    "ocean_server_ssl_enabled",
]


class Settings(BaseSettings):
    environment: str = ""

    ocean_db_impl: str = "duckdb"
    ocean_api_impl: str = "local"

    clickhouse_host: str = None # type: ignore
    clickhouse_port: str = None

    persist_directory: str = ".ocean"

    ocean_server_host: str = None
    ocean_server_http_port: str = None
    ocean_server_ssl_enabled: bool = False
    ocean_server_grpc_port: str = None
    ocean_server_cors_allow_origins: List[str] = []  # eg ["http://localhost:3000"]

    anonymized_telemetry: bool = True

    def __getitem__(self, item):
        return getattr(self, item)

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"
